from .BrentMet_var2_help import *
import numpy as np
from re import findall
from re import sub
import pandas as pd
from sympy import *


class AdjointGrad:
    def find(z, x, e, flag1, flag2, extr):

        """
        Поиск экстремума функции многих переменных при помощи алгоритма Ньютона-сопряженного градиента
        Parameters
        ===========
        :param z: str
            функция минимизации (максимизации)
        :param x: float
            начальная точка, из которой начинаем спуск
        :param e: float
            точность оптимизации
        :param flag1: int
            Вывод промежуточных результатов? 1 - да / 0 - нет
        :param flag2: int
            запись промежуточных результатов в датасет 1 - да / 0 - нет
        :param extr: int
            экстремум который вы хотите найти 1 - максимум, 0 - минимум
        :return:
            Возвращает точку максимума или минимума, значение функции в точке минимума и датасет с промежуточными
             вычислениями
        """

        df = pd.DataFrame(columns=['Номер итерации', 'Полученные значения'])
        c=0
        func = sympify(z)
        l = Symbol('l')
        lst_xi = np.sort(list(set(findall(r'[x]\d', z))))

        grad = []
        for i in range(len(lst_xi)):
            grad.append(func.diff(lst_xi[i]))
        x = np.array(x)
        j = 0
        while True:
            i = 0
            su = dict(zip(lst_xi, x))
            calc_grad = []
            for k in range(len(lst_xi)):
                calc_grad.append(grad[k].subs(su))
            calc_grad = np.array(calc_grad)
            S = -1*calc_grad
            x_jk = x

            while i+1<len(lst_xi):

                lam = x + l*S
                su1 = dict(zip(lst_xi, lam))
                f_l = str(func.subs(su1))
                f_l = sub(r'l', r'x', f_l)
                lam = BrentMet()
                
                l_min = lam.find(f_l, -100, 100, e*10, 100)
                
                x_jk_1 = x_jk + l_min*S


                su_xjk_1 = dict(zip(lst_xi, x_jk_1))
                calc_grad_xjk_1 = []
                for k in range(len(lst_xi)):
                    calc_grad_xjk_1.append(grad[k].subs(su_xjk_1))
                calc_grad_xjk_1 = np.array(calc_grad_xjk_1)

                sum_grad = (calc_grad**2).sum()
                sum_grad_k = (calc_grad_xjk_1**2).sum()
                Skj = -1*calc_grad_xjk_1 + (sum_grad_k/sum_grad) * S

                met_Skj = (Skj**2).sum()**(1/2)
                su_after = su_xjk_1

                x_beauty = x_jk_1
                if flag1 == 1:
                    print(f'Итерация №{i}, точка {tuple(x_beauty)}')
                if flag2 == 1:
                    df.loc[i] = [i, tuple(x_beauty)]

                x_met = ((x_jk_1 - x_jk)**2).sum()**(1/2)
                # Написать датасеты использовать x_beauty
                if extr == 0:

                    if (met_Skj < e) or (x_met < e):
                        print(f'Минимум в точке {tuple(x_beauty)}')
                        print(f'Значение функции в точке минимума {float(func.subs(su_after))}')
                        c = 1
                        break
                else:
                    if (met_Skj < e) or (x_met < e):
                        print(f'Максимум в точке {tuple(x_beauty)}')
                        print(f'Значение функции в точке максимума {float(func.subs(su_after))}')
                        c = 1
                        break
                x_jk = x_jk_1
                S = Skj
                i += 1
            if c == 1:
                break
            x = x_jk
            j += 1
        return x_beauty

#function = AdjointGrad()
#function.find('3.125*(-0.2*x1 - 0.6*x2 - x3 + 0.6)**2 + 4.5*(-0.166666666666667*x1 - 0.333333333333333*x2 - x3 + 0.666666666666667)**2 + 6.125*(-0.142857142857143*x1 - 0.571428571428571*x2 - x3 + 0.857142857142857)**2 + 8.0*(-0.125*x1 - 0.125*x2 - x3 + 0.625)**2', [-3,1,1], 0.0001, 0, 0, 0)



import numpy as np
from typing import Callable, Optional
# import autograd.numpy as npa
from autograd import grad, hessian
from matplotlib import pyplot as plt
import numexpr as ne
import sympy as sp
import copy

from .input_validation import check_expression, check_restr, check_point
from .prepocessing import prepare_all


class Newton:
    """
    Метод решения задачи методом Ньютона.

    Parameters
    ----------
    function: Callable
        Функция для минимизации.

    restrictions: list
        Список вызываемых функций питона, которые имеют смысл левой части ограничений g(x) = 0

    x0: np.ndarray
        Плоский массив, содержащий начальную точку.

    eps: Optional[float] = 1e-12
        Точность. От этого параметра зависит, когда будет остановлен итерационные процесс поиска решения. При
        слишком малых значениях, есть вероятность бесконечного цикла или nan в ответе.
    """

    def __init__(self,
                 function: Callable,
                 restrictions: list,
                 x0: np.ndarray,
                 eps: Optional[float] = 1e-12):
        self.function = function
        self.restrictions = restrictions
        self.x = x0
        self.eps = eps
        self.mu = 1

    def solve(self):
        """
        Метод решает задачу.

        Returns
        -------
        x: np.ndarray
            Массив с решениями прямой задачи.

        lam: np.ndarray
            Массив с решениями двойственной задачи.
        """

        hess_lagrange = hessian(self.lagrange)
        grad_lagrange = grad(self.lagrange)
        step = np.linalg.inv(hess_lagrange(self.x)) @ grad_lagrange(self.x)
        while not np.allclose(step, np.zeros(self.x.shape)):
            self.x = self.x - step
            step = np.linalg.inv(hess_lagrange(self.x)) @ grad_lagrange(self.x)
            self.mu *= 10
        return self.x

    def lagrange(self, x):
        L = self.function(x)
        k = 0
        for i in self.restrictions:
            k += self.mu * (i(x)) ** 2
        L += k
        return L


if __name__ == '__main__':
    f = 'x1**2 + x2**2 + (0.5*1*x1 + 0.5*2*x2)**2 + (0.5*1*x1 + 0.5*2*x2)**4'
    f_graph = copy.deepcopy(f)
    subject_to = 'x1+x2=0;2*x1-3*x2=0'
    # point_min = np.array([0, 0])
    # point_start = np.array([-5, 4.])
    point_start = '-5;4'

    # input_validation
    f = check_expression(f)
    subject_to = check_restr(subject_to, method='Newton')
    point_start = check_point(point_start, f, subject_to, 'Newton')
    # preprocessing
    f, subject_to, point_start = prepare_all(f, subject_to, 'Newton', point_start)
    # solver
    task = Newton(f, subject_to, point_start)
    ans = task.solve()
    # print(np.allclose(ans, point_min))
    print(ans)

    # Рисуем график
    fig = plt.figure(figsize=(10, 8))
    ax = fig.add_subplot(111, projection='3d')
    ax.set_title('Метод Ньютона')
    ax.set_zlabel('F(x1,x2)')
    ax.set_ylabel('x2')
    ax.set_xlabel('x1')

    x1 = np.linspace(-1, 1, 25)
    x2 = np.linspace(-1, 1, 25)
    x1, x2 = np.meshgrid(x1, x2)
    Z = ne.evaluate(f_graph)

    surf = ax.plot_surface(x1, x2, Z, cmap='viridis', shade=True, alpha=0.5)
    fig.colorbar(surf, shrink=0.5, aspect=10)

    func1 = sp.sympify(f_graph)

    x1, x2 = sp.symbols('x1 x2')

    ax.scatter(float(ans[0]), float(ans[1]),
               func1.subs([(x1, float(ans[0])), (x2, float(ans[1]))]),
               c='r', s=50,
               label=f'Точка экстремума \n x1 = {float(ans[0])},\n x2 = {float(ans[1])},\n'
                     f' F(x1,x2) = {func1.subs([(x1, float(ans[0])), (x2, float(ans[1]))])}')

    ax.legend(labelcolor='black')
    plt.show()



import matplotlib.pyplot as plt
import numexpr as ne
import numpy as np
import matplotlib.animation as animation
import pandas as pd
from sympy import *
from re import findall
from re import sub

from .BrentMet_var2_help import *



class FastGradDesent:
    def find(z, x, iter, e, flag1, flag2, extr, namegraph):
        """
        Поиск экстремума функции многих переменных методом наискорейшего градиентного спуска
        Parameters
        ===========
        :param z: str
            функция минимизации (максимизации)
        :param x: float
            начальная точка, из которой начинаем спуск
        :param iter: int
            число итераций
        :param e: float
            точность оптимизации
        :param flag1: int
            Вывод промежуточных результатов? 1 - да / 0 - нет
        :param flag2: int
            запись промежуточных результатов в датасет 1 - да / 0 - нет
        :param extr: int
            экстремум который вы хотите найти 1 - максимум, 0 - минимум
        :param namegraph: str
            принимает параметр str, рисует график только 2д, тоесть функции только одной переменной
        :return:
            Возвращает точку максимума или минимума, значение функции в точке минимума и датасет с промежуточными
             вычислениями
        """

        func = sympify(z)
        lst_x_plot = []
        for i in range(len(x)):
            lst_x_plot.append(float(x[i][0]))
        lst_xi = np.sort(list(set(findall(r'[x]\d', z))))
        l = Symbol('l')
        det_func = []
        for i in range(len(lst_xi)):
            det_func.append(func.diff(lst_xi[i]))
        i = 0

        df = pd.DataFrame(columns=['Номер итерации', 'Полученные значения'])

        while i < iter:
            podst = []
            calc_det = []
            for j in range(len(lst_xi)):
                podst.append((lst_xi[j], x[j][0]))
            for j in range(len(lst_xi)):
                calc_det.append([det_func[j].subs(podst)])

            lam = x - np.dot(l, calc_det)

            podst_f_l = []
            for j in range(len(lam)):
                podst_f_l.append((lst_xi[j], lam[j][0]))
            f_l = str(func.subs(podst_f_l))
            f_l = sub(r'l', r'x', f_l)

            lam = BrentMet()
            l_min = lam.find(f_l, -5, 5, e, iter)
            xk = x - np.dot(l_min, calc_det)

            podst_after = []
            for j in range(len(lst_xi)):
                podst_after.append((lst_xi[j], xk[j][0]))
            # Написать датасеты использовать x_beauty
            x = xk
            x_beauty = []
            for j in range(len(x)):
                x_beauty.append(float(x[j][0]))
            if flag1 == 1:
                print(f'Итерация №{i}, точка {tuple(x_beauty)}')
            if flag2 == 1:
                df.loc[i] = [i, tuple(x_beauty)]

            for j in range(len(x)):
                lst_x_plot.append(float(x[j][0]))

            if extr == 0:
                if func.subs(podst_after) > func.subs(podst):
                    print(f'Минимум в точке {tuple(x_beauty)}')
                    print(f'Значение функции в точке минимума {float(func.subs(podst))}')
                    print(f'Неточное решение. При следующем запуске уменьшите лямбду.')
                    break
                elif abs(func.subs(podst) - func.subs(podst_after)) < e:
                    print(f'Минимум в точке {tuple(x_beauty)}')
                    print(f'Значение функции в точке минимума {float(func.subs(podst_after))}')
                    break

                elif i == (iter - 1):
                    print(f'Неточное решение. На последней итерации минимум в точке {tuple(x_beauty)}')
                    print(f'Значение функции в точке минимума {float(func.subs(podst_after))}')
                    break
            else:
                if func.subs(podst_after) > func.subs(podst):
                    print(f'Максимум в точке {tuple(x_beauty)}')
                    print(f'Значение функции в точке максимума {float(func.subs(podst))}')
                    print(f'Неточное решение. При следующем запуске уменьшите лямбду.')
                    break
                elif abs(func.subs(podst) - func.subs(podst_after)) < e:
                    print(f'Максимум в точке {tuple(x_beauty)}')
                    print(f'Значение функции в точке максимума {float(func.subs(podst_after))}')
                    break

                elif i == (iter - 1):
                    print(f'Неточное решение. На последней итерации максимум в точке {tuple(x_beauty)}')
                    print(f'Значение функции в точке максимума {float(func.subs(podst_after))}')
                    break

            i += 1

        if flag2 == 1:
            print(df)
            df.to_csv('FastGradDescent_results.csv', index=False)
            print('Датасет сохранен в папку проекта в формате csv')

        if len(x_beauty) == 1:
            fig, ax = plt.subplots()

            ax.set_title('Поиск экстремумов')
            ax.set_ylabel('Y')
            ax.set_xlabel('X')

            x1 = np.linspace(-10, 10, 100)
            y = ne.evaluate(z)

            plt.plot(x1, y, color='b')

            redDot, = plt.plot([], [], 'ro')

            def animate(x1):
                redDot.set_data(x1, ne.evaluate(z))
                return redDot,

            myAnimation = animation.FuncAnimation(fig, animate, frames=lst_x_plot,
                                                  interval=500, blit=True, repeat=True)
            plt.show()
        
            myAnimation.save(f'{namegraph}.gif', writer='pillow')


# while True:
#     print("Какой экстремум вы хотите найти? 1 - максимум \ 0 - минимум")
#     q = int(input())
#     if q == 1:
#         extr = 1
#         break
#     elif q == 0:
#         extr = 0
#         break
#     else:
#         print("Такой команды нет. Введите снова")
# print("Введите функцию f(x1, x2, ... , xn). Например: x1**2 + x1*x2**3 + x3**2")
# f = input()
# if extr == 1:
#     f = "- (" + f + ")"
#     print(f)
# count_param = np.sort(list(set(re.findall(r'[x]\d', f))))
# print("Введите начальную точку, из которой начинаем спуск.")
# lst_xi = []
# for i in range(len(count_param)):
#     print(f'Введите {count_param[i]}. Например: 1')
#     xi = [float(input())]
#     lst_xi.append(xi)
# while True:
#     print("Хотите ввести число итераций? 1 - да / 0 - нет")
#     q = int(input())
#     if q == 1:
#         print("Введите кол-во итераций. Например: 500")
#         iter = int(input())
#         break
#     elif q == 0:
#         iter = 500
#         break
#     else:
#         print("Такой команды нет. Введите снова")
# while True:
#     print("Хотите ввести точность оптимизации? 1 - да / 0 - нет")
#     q = int(input())
#     if q == 1:
#         print("Введите точность оптимизации. Например: 0.001")
#         e = float(input())
#         break
#     elif q == 0:
#         e = 0.0001
#         break
#     else:
#         print("Такой команды нет. Введите снова")
#
# while True:
#     print("Хотите ввести промежуточные результаты? 1 - да / 0 - нет")
#     q = int(input())
#     if q == 1:
#         flag1 = 1
#         break
#     elif q == 0:
#         flag1 = 0
#         break
#     else:
#         print("Такой команды нет. Введите снова")
#
# while True:
#     print("Хотите записать промежуточные результаты в датасет? 1 - да / 0 - нет")
#     q = int(input())
#     if q == 1:
#         flag2 = 1
#         break
#     elif q == 0:
#         flag2 = 0
#         break
#     else:
#         print("Такой команды нет. Введите снова")
#
# function = FastGradDesent()
# function.find(f, lst_xi, iter, e, flag1, flag2, extr)

# function.find("x1**2+x2", [[1], [5]], 100, 0.001, 0, 0, 0)
